@@ -61,6 +61,12 @@ goa_ews_client_new (void)
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+typedef struct
+{
+  gchar *password;
+  gchar *username;
+} AutodiscoverAuthData;
+
 typedef struct
 {
   GCancellable *cancellable;
@@ -71,13 +77,16 @@ typedef struct
   guint pending;
   gulong cancellable_id;
   xmlOutputBuffer *buf;
+  AutodiscoverAuthData *auth;
 } AutodiscoverData;
 
-typedef struct
+static void
+ews_client_autodiscover_auth_data_free (AutodiscoverAuthData *auth)
 {
-  gchar *password;
-  gchar *username;
-} AutodiscoverAuthData;
+  g_free (auth->password);
+  g_free (auth->username);
+  g_slice_free (AutodiscoverAuthData, auth);
+}
 
 static void
 ews_client_autodiscover_data_free (gpointer user_data)
@@ -92,22 +101,12 @@ ews_client_autodiscover_data_free (gpointer user_data)
 
   g_clear_error (&data->error);
 
-  /* soup_session_queue_message stole the references to data->msgs */
   xmlOutputBufferClose (data->buf);
+  g_clear_pointer (&data->auth, ews_client_autodiscover_auth_data_free);
   g_object_unref (data->session);
   g_slice_free (AutodiscoverData, data);
 }
 
-static void
-ews_client_autodiscover_auth_data_free (gpointer data, GClosure *closure)
-{
-  AutodiscoverAuthData *auth = data;
-
-  g_free (auth->password);
-  g_free (auth->username);
-  g_slice_free (AutodiscoverAuthData, auth);
-}
-
 static gboolean
 ews_client_check_node (const xmlNode *node, const gchar *name)
 {
@@ -115,9 +114,8 @@ ews_client_check_node (const xmlNode *node, const gchar *name)
   return node->type == XML_ELEMENT_NODE && !g_strcmp0 ((gchar *) node->name, name);
 }
 
-static void
-ews_client_authenticate (SoupSession *session,
-                         SoupMessage *msg,
+static gboolean
+ews_client_authenticate (SoupMessage *msg,
                          SoupAuth *auth,
                          gboolean retrying,
                          gpointer user_data)
@@ -125,26 +123,26 @@ ews_client_authenticate (SoupSession *session,
   AutodiscoverAuthData *data = user_data;
 
   if (retrying)
-    return;
+    return FALSE;
 
   soup_auth_authenticate (auth, data->username, data->password);
+  return TRUE;
 }
 
-static void
-ews_client_request_started (SoupSession *session, SoupMessage *msg, SoupSocket *socket, gpointer user_data)
+static gboolean
+ews_client_accept_certificate (SoupMessage *msg, GTlsCertificate *cert, GTlsCertificateFlags cert_flags, gpointer user_data)
 {
   AutodiscoverData *data;
   GTask *task = G_TASK (user_data);
-  GTlsCertificateFlags cert_flags;
 
-  g_debug ("goa_ews_client_autodiscover(): request started (%p)", msg);
+  g_debug ("goa_ews_client_autodiscover(): accept certificate for request (%p)", msg);
 
   data = (AutodiscoverData *) g_task_get_task_data (task);
 
-  if (!data->accept_ssl_errors
-      && soup_message_get_https_status (msg, NULL, &cert_flags)
-      && cert_flags != 0
-      && data->error == NULL)
+  if (data->accept_ssl_errors || cert_flags == 0)
+    return TRUE;
+
+  if (data->error == NULL)
     {
       goa_utils_set_error_ssl (&data->error, cert_flags);
 
@@ -153,6 +151,8 @@ ews_client_request_started (SoupSession *session, SoupMessage *msg, SoupSocket *
        */
       soup_session_abort (data->session);
     }
+
+  return FALSE;
 }
 
 static void
@@ -191,9 +191,25 @@ ews_client_autodiscover_parse_protocol (xmlNode *node)
   return as_url && oab_url;
 }
 
-static void
-ews_client_autodiscover_response_cb (SoupSession *session, SoupMessage *msg, gpointer user_data)
+static gboolean
+ews_client_autodiscover_complete_idle_cb (gpointer user_data)
 {
+  GTask *task = user_data;
+  AutodiscoverData *data = g_task_get_task_data (task);
+
+  if (data->error != NULL)
+    g_task_return_error (task, g_steal_pointer (&data->error));
+  else
+    g_task_return_boolean (task, TRUE);
+
+  return G_SOURCE_REMOVE;
+}
+
+static void
+ews_client_autodiscover_response_cb (SoupSession *session, GAsyncResult *result, gpointer user_data)
+{
+  SoupMessage *msg;
+  GBytes *body;
   GError *error = NULL;
   AutodiscoverData *data;
   GTask *task = G_TASK (user_data);
@@ -204,7 +220,11 @@ ews_client_autodiscover_response_cb (SoupSession *session, SoupMessage *msg, gpo
   xmlDoc *doc;
   xmlNode *node;
 
-  g_debug ("goa_ews_client_autodiscover(): response (%p, %u)", msg, msg->status_code);
+  msg = soup_session_get_async_result_message (session, result);
+
+  g_debug ("goa_ews_client_autodiscover(): response (%p, %u)", msg, soup_message_get_status (msg));
+
+  body = soup_session_send_and_read_finish (session, result, &error);
 
   data = (AutodiscoverData *) g_task_get_task_data (task);
   size = sizeof (data->msgs) / sizeof (data->msgs[0]);
@@ -215,16 +235,19 @@ ews_client_autodiscover_response_cb (SoupSession *session, SoupMessage *msg, gpo
         break;
     }
   if (idx == size || data->pending == 0)
-    return;
+    {
+      g_bytes_unref (body);
+      g_clear_object (&error);
+      g_object_unref (task);
+      return;
+    }
 
   data->msgs[idx] = NULL;
-  status = msg->status_code;
-
-  /* status == SOUP_STATUS_CANCELLED, if we are being aborted by the
+  /* G_IO_ERROR_CANCELLED, if we are being aborted by the
    * GCancellable, an SSL error or another message that was
    * successful.
    */
-  if (status == SOUP_STATUS_CANCELLED)
+  if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
     {
       /* If we are being aborted by the GCancellable, then the
        * GTask is responsible for setting the GError automatically.
@@ -235,21 +258,23 @@ ews_client_autodiscover_response_cb (SoupSession *session, SoupMessage *msg, gpo
        */
       goto out;
     }
-  else if (status != SOUP_STATUS_OK)
+
+  status = soup_message_get_status (msg);
+  if (status != SOUP_STATUS_OK || error)
     {
-      g_warning ("goa_ews_client_autodiscover() failed: %u — %s", msg->status_code, msg->reason_phrase);
+      g_warning ("goa_ews_client_autodiscover() failed: %u — %s", status, soup_message_get_reason_phrase (msg));
       g_return_if_fail (data->error == NULL);
 
-      goa_utils_set_error_soup (&error, msg);
+      if (!error)
+        goa_utils_set_error_soup (&error, msg);
       goto out;
     }
 
-  soup_buffer_free (soup_message_body_flatten (SOUP_MESSAGE (msg)->response_body));
   g_debug ("The response headers");
   g_debug ("===================");
-  g_debug ("%s", SOUP_MESSAGE (msg)->response_body->data);
+  g_debug ("%s", (char *)g_bytes_get_data (body, NULL));
 
-  doc = xmlReadMemory (msg->response_body->data, msg->response_body->length, "autodiscover.xml", NULL, 0);
+  doc = xmlReadMemory (g_bytes_get_data (body, NULL), g_bytes_get_size (body), "autodiscover.xml", NULL, 0);
   if (doc == NULL)
     {
       g_set_error (&error,
@@ -333,7 +358,7 @@ ews_client_autodiscover_response_cb (SoupSession *session, SoupMessage *msg, gpo
           /* The callback (ie. this function) will be invoked after we
            * have returned to the main loop.
            */
-          soup_session_cancel_message (data->session, data->msgs[idx], SOUP_STATUS_CANCELLED);
+          g_cancellable_cancel (data->cancellable);
         }
     }
 
@@ -360,10 +385,10 @@ ews_client_autodiscover_response_cb (SoupSession *session, SoupMessage *msg, gpo
   data->pending--;
   if (data->pending == 0)
     {
-      if (data->error != NULL)
-        g_task_return_error (task, g_steal_pointer (&data->error));
-      else
-        g_task_return_boolean (task, TRUE);
+      /* This is required, to not free the SoupSession under its hands (as this callback is called under it) */
+      GSource *idle_source = g_idle_source_new ();
+      g_source_set_callback (idle_source, ews_client_autodiscover_complete_idle_cb, g_object_ref (task), g_object_unref);
+      g_source_attach (idle_source, g_task_get_context (task));
     }
 
   g_clear_error (&error);
@@ -399,52 +424,56 @@ static void
 ews_client_post_restarted_cb (SoupMessage *msg, gpointer data)
 {
   xmlOutputBuffer *buf = data;
+  GBytes *body;
 
   /* In violation of RFC2616, libsoup will change a POST request to
    * a GET on receiving a 302 redirect.
    */
   g_debug ("Working around libsoup bug with redirect");
-  g_object_set (msg, SOUP_MESSAGE_METHOD, "POST", NULL);
+  g_object_set (msg, "method", "POST", NULL);
 
-  soup_message_set_request(msg,
-                           "text/xml; charset=utf-8",
-                           SOUP_MEMORY_COPY,
 #ifdef LIBXML2_NEW_BUFFER
-                           (gchar *) xmlOutputBufferGetContent(buf),
-                           xmlOutputBufferGetSize(buf));
+  body = g_bytes_new (xmlOutputBufferGetContent (buf), xmlOutputBufferGetSize (buf));
 #else
-                           (gchar *) buf->buffer->content,
-                           buf->buffer->use);
+  body = g_bytes_new (buf->buffer->content, buf->buffer->use);
 #endif
+  soup_message_set_request_body_from_bytes (msg, "text/xml; charset=utf-8", body);
+  g_bytes_unref (body);
 }
 
 static SoupMessage *
-ews_client_create_msg_for_url (const gchar *url, xmlOutputBuffer *buf)
+ews_client_create_msg_for_url (const gchar *url, xmlOutputBuffer *buf, AutodiscoverAuthData *auth, GTask *task)
 {
   SoupMessage *msg;
+  GBytes *body = NULL;
 
   msg = soup_message_new (buf != NULL ? "POST" : "GET", url);
-  soup_message_headers_append (msg->request_headers, "User-Agent", "libews/0.1");
+  soup_message_headers_append (soup_message_get_request_headers (msg),
+                               "User-Agent", "libews/0.1");
+
+  g_signal_connect (msg, "authenticate",
+                    G_CALLBACK (ews_client_authenticate),
+                    auth);
+  g_signal_connect (msg, "accept-certificate",
+                    G_CALLBACK (ews_client_accept_certificate),
+                    task);
 
   if (buf != NULL)
     {
-      soup_message_set_request (msg,
-                                "text/xml; charset=utf-8",
-                                SOUP_MEMORY_COPY,
 #ifdef LIBXML2_NEW_BUFFER
-                                (gchar *) xmlOutputBufferGetContent(buf),
-                                xmlOutputBufferGetSize(buf));
+      body = g_bytes_new (xmlOutputBufferGetContent (buf), xmlOutputBufferGetSize (buf));
 #else
-                                (gchar *) buf->buffer->content,
-                                buf->buffer->use);
+      body = g_bytes_new (buf->buffer->content, buf->buffer->use);
 #endif
+      soup_message_set_request_body_from_bytes (msg, "text/xml; charset=utf-8", body);
       g_signal_connect (msg, "restarted", G_CALLBACK (ews_client_post_restarted_cb), buf);
     }
 
-  soup_buffer_free (soup_message_body_flatten (SOUP_MESSAGE (msg)->request_body));
   g_debug ("The request headers");
   g_debug ("===================");
-  g_debug ("%s", SOUP_MESSAGE (msg)->request_body->data);
+  g_debug ("%*.s", body ? (int) g_bytes_get_size (body) : 0, body ? (char *)g_bytes_get_data (body, NULL) : "");
+  if (body)
+    g_bytes_unref (body);
 
   return msg;
 }
@@ -497,12 +526,15 @@ goa_ews_client_autodiscover (GoaEwsClient        *self,
    * (successful) one win.
    */
 
+  auth = g_slice_new0 (AutodiscoverAuthData);
+  auth->username = g_strdup (username);
+  auth->password = g_strdup (password);
+  data->auth = auth;
   data->buf = buf;
-  data->msgs[0] = ews_client_create_msg_for_url (url1, buf);
-  data->msgs[1] = ews_client_create_msg_for_url (url2, buf);
+  data->msgs[0] = ews_client_create_msg_for_url (url1, buf, auth, task);
+  data->msgs[1] = ews_client_create_msg_for_url (url2, buf, auth, task);
   data->pending = sizeof (data->msgs) / sizeof (data->msgs[0]);
-  data->session = soup_session_new_with_options (SOUP_SESSION_SSL_STRICT, FALSE,
-                                                 NULL);
+  data->session = soup_session_new ();
   soup_session_add_feature_by_type (data->session, SOUP_TYPE_AUTH_NTLM);
   data->accept_ssl_errors = accept_ssl_errors;
 
@@ -515,26 +547,18 @@ goa_ews_client_autodiscover (GoaEwsClient        *self,
                                                     NULL);
     }
 
-  auth = g_slice_new0 (AutodiscoverAuthData);
-  auth->username = g_strdup (username);
-  auth->password = g_strdup (password);
-  g_signal_connect_data (data->session,
-                         "authenticate",
-                         G_CALLBACK (ews_client_authenticate),
-                         auth,
-                         ews_client_autodiscover_auth_data_free,
-                         0);
-
-  g_signal_connect (data->session, "request-started", G_CALLBACK (ews_client_request_started), task);
-
-  soup_session_queue_message (data->session,
-                              data->msgs[0],
-                              ews_client_autodiscover_response_cb,
-                              g_object_ref (task));
-  soup_session_queue_message (data->session,
-                              data->msgs[1],
-                              ews_client_autodiscover_response_cb,
-                              g_object_ref (task));
+  soup_session_send_and_read_async (data->session,
+                                    data->msgs[0],
+                                    G_PRIORITY_DEFAULT,
+                                    data->cancellable,
+                                    (GAsyncReadyCallback)ews_client_autodiscover_response_cb,
+                                    g_object_ref (task));
+  soup_session_send_and_read_async (data->session,
+                                    data->msgs[1],
+                                    G_PRIORITY_DEFAULT,
+                                    data->cancellable,
+                                    (GAsyncReadyCallback)ews_client_autodiscover_response_cb,
+                                    g_object_ref (task));
 
   g_free (url2);
   g_free (url1);
