@@ -1097,7 +1097,7 @@ on_identity_added (GoaIdentityManager *identity_manager,
 
   object = find_object_with_principal (self, identifier, FALSE);
 
-  if (object == NULL)
+  if (object == NULL && goa_identity_is_signed_in (identity))
     add_temporary_account (self, identity);
 
   g_clear_object (&object);
@@ -1469,11 +1469,9 @@ on_identity_expiring (GoaIdentityManager *identity_manager,
   ensure_account_credentials (self, object);
   g_clear_object (&object);
 }
-
 static void
-on_identity_expired (GoaIdentityManager *identity_manager,
-                     GoaIdentity        *identity,
-                     GoaIdentityService *self)
+handle_identity_expired (GoaIdentityService *self,
+                         GoaIdentity        *identity)
 {
   const char *principal;
   GoaObject  *object;
@@ -1491,6 +1489,14 @@ on_identity_expired (GoaIdentityManager *identity_manager,
   g_clear_object (&object);
 }
 
+static void
+on_identity_expired (GoaIdentityManager *identity_manager,
+                     GoaIdentity        *identity,
+                     GoaIdentityService *self)
+{
+  handle_identity_expired (self, identity);
+}
+
 static void
 on_sign_out_for_account_change_done (GoaIdentityService *self,
                                      GAsyncResult       *result)
@@ -1705,6 +1711,12 @@ on_identities_listed (GoaIdentityManager *manager,
         g_object_unref (object);
 
       g_free (object_path);
+
+      /* Treat identities that started out expired as if they just expired, in case
+       * the identity service is started long after goa-daemon
+       */
+      if (!goa_identity_is_signed_in (identity))
+        handle_identity_expired (self, identity);
     }
 
  out:
